/*     DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
                    Version 2, December 2004

 Copyright (C) 2009 Adinpsz contributors
  7 chemin Villebois 38100 Grenoble
 Everyone is permitted to copy and distribute verbatim or modified
 copies of this license document, and changing it is allowed as long
 as the name is changed.

            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  0. You just DO WHAT THE FUCK YOU WANT TO. */
/* This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details. */ 
module audio.audiodata;

private import audio.audiomodule;
private import audio.types;
private import audio.audiobuffer;
private import audio.constants;
private import audio.message;


class AudioData // audio buffer
{
	private
	{
		AudioModule m_writer;	
		AudioCycle m_lastCycle;
		AudioDataType m_type;
	}
	
	public
	{
		this(AudioDataType type, AudioModule writer)
		{
			m_writer = writer;
			m_lastCycle = -1;
			m_type = type;
		}
		
		AudioModule writer()
		{
			return m_writer;
		}		
		
		void updateTo(AudioCycle currentCycle)
		{
			if (lastCycle < currentCycle)
			{
				writer.updateTo(currentCycle);
				lastCycle = currentCycle; // to prevent cycles
			}	
		}
		
		AudioCycle lastCycle(AudioCycle last)
		{
			return m_lastCycle = last;
		}
		
		AudioCycle lastCycle()
		{
			return m_lastCycle;
		}
		
		AudioDataType type()
		{
			return m_type;	
		}
		
		float asFloat()
		{
			return (cast(FloatData)this).value;
		}
		
		int asInt()
		{
			return (cast(IntegerData)this).value;
		}
		
		float * asBuffer()
		{
			return (cast(AnalogData)this).buffer().ptr;
		}
		
		float * asLeftBuffer()
		{
			return (cast(StereoAnalogData)this).left().ptr;
		}
		
		float * asRightBuffer()
		{
			return (cast(StereoAnalogData)this).left().ptr;
		}
		
		
		
	}
}

final class FloatData : AudioData
{
	private
	{
		float m_value;	
	}
	
	public
	{
		this(AudioModule writer)
		{
			super(AudioDataType.FLOAT, writer);
		}		
		
		void value(float _value)
		{
			m_value = _value;
		}	
		
		float value()
		{
			return m_value;
		}
		
	}
}

final class AnalogData : AudioData
{
	private
	{
		AudioBuffer m_buffer;
	}
	
	public
	{
		this(AudioModule writer)
		{
			super(AudioDataType.ANALOG, writer);
			m_buffer = new AudioBuffer(BUFFER_SIZE);			
		}
		
		AudioBuffer buffer()
		{
			return m_buffer;
		}
	}
}

final class StereoAnalogData : AudioData
{
	private
	{
		AudioBuffer m_left;
		AudioBuffer m_right;
	}
	
	public
	{
		this(AudioModule writer)
		{
			super(AudioDataType.STEREO_ANALOG, writer);
			m_left = new AudioBuffer(BUFFER_SIZE);
			m_right = new AudioBuffer(BUFFER_SIZE);
		}
		
		AudioBuffer left()
		{
			return m_left;
		}	
		
		AudioBuffer right()
		{
			return m_right;
		}		
	}
}

final class IntegerData : AudioData
{
	private
	{
		int m_value;	
	}
	
	public
	{
		this(AudioModule writer)
		{
			super(AudioDataType.INTEGER, writer);
		}		
		
		int value(int _value)
		{
			return m_value = _value;
		}	
		
		int value()
		{
			return m_value;
		}		
	}
}

final class MessageData : AudioData
{
	private
	{
		const int MAX_MESSAGES = 16;
		const int MESSAGES_MASK = MAX_MESSAGES - 1;
		
		Message[MAX_MESSAGES] m_messages;	
		
		int m_read, m_write;
		
		int next(int index)
		{
			return (index + 1) & MESSAGES_MASK;
		}
	}
	
	public
	{
		this(AudioModule writer)
		{
			super(AudioDataType.MESSAGE, writer);
			m_read = 0;
			m_write = 0;
		}		
		
		int opApply(int delegate(inout Message) dg) 
		{	
			int result = 0;
			for (int i = m_read; i != m_write; i = next(i))
			{
				result = dg(m_messages[i]);	
			}
			return result;			
		}
		
		void post(Message m)
		{
			auto n = next(m_write);
			if ( n == m_read) // full
			{
				return; // do nothing
			}
			
			m_messages[m_write] = m;
			m_write = n;
		}	
		
		int count()
		{
			return (MAX_MESSAGES + m_write - m_read) & MESSAGES_MASK;
		}
		
		void clear()
		{
			m_write = m_read;
		}
		
	}
}

final class AudioDataFactory
{
	static AudioData create(AudioDataType t, AudioModule writer)
	{
		switch (t)
		{
			case AudioDataType.ANALOG:
				return new AnalogData(writer);
				
			case AudioDataType.STEREO_ANALOG:
				return new StereoAnalogData(writer);
			
			case AudioDataType.FLOAT:
				return new FloatData(writer);
			
			case AudioDataType.INTEGER:
				return new IntegerData(writer);
			
			case AudioDataType.MESSAGE:
				return new MessageData(writer);
				
			default: ;
		}
		return null;
	}
}