/*     DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
                    Version 2, December 2004

 Copyright (C) 2009 Adinpsz contributors
  7 chemin Villebois 38100 Grenoble
 Everyone is permitted to copy and distribute verbatim or modified
 copies of this license document, and changing it is allowed as long
 as the name is changed.

            DO WHAT THE FUCK YOU WANT TO PUBLIC LICENSE
   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION

  0. You just DO WHAT THE FUCK YOU WANT TO. */
/* This program is free software. It comes without any warranty, to
 * the extent permitted by applicable law. You can redistribute it
 * and/or modify it under the terms of the Do What The Fuck You Want
 * To Public License, Version 2, as published by Sam Hocevar. See
 * http://sam.zoy.org/wtfpl/COPYING for more details. */ 
module audio.pin;


private import audio.types;
private import audio.audiodata;
private import audio.audiomodule;
private import audio.modules.smoother;
private import audio.modules.doubler;
private import audio.modules.constant;
private import audio.errors;

class Pin
{
	protected
	{
		AudioData m_data;
		char[] m_name;	
		AudioModule m_owner;		
		Pin m_linkedPin; // a pin can be "linked" to another pin (link !is null). "linking" is delegation.
		
		this(AudioModule owner, char[] name, AudioData data)
		{
			m_name = name;
			m_data = data;
			m_owner = owner;
		}	
		
		// return true if this could be "linked" to pin 				
		abstract bool canBeLinkedTo(Pin pin);
		
	}
	
	public
	{
		final AudioData data()
		{
			if (m_linkedPin is null)
			{
				return m_data;	
			} else
			{
				return m_linkedPin.data();	// recursive call to "linked" pin
			}
		}
		
		
		final char[] name()
		{
			return m_name;	
		}
		
		final bool assigned()
		{
			return data() !is null;	
		}
		
		final bool unassigned()
		{
			return data() is null;	
		}
		
		final AudioModule owner()
		{
			return m_owner; // owner do not change
		}
		
		final void link(Pin newLink) // link a pin to another pin
		{
			if (canBeLinkedTo(newLink))
			{
				this.m_linkedPin = newLink;
			} else
			{
				throw new LinkIsImpossible(this, newLink);
			}
		}
		
		final void unlink() // erase a link
		{
			if (m_linkedPin !is null)
			{
				m_linkedPin = null;	
			}
		}
	}
}

final class InputPin : Pin
{
	private
	{
		AudioDataTypeSet m_expectedType;
		
		void setData(AudioData audioData)
		{
			auto type = audioData.type();
			
			if (type & m_expectedType)
			{
				m_data = audioData; // OK
			} else
			{
				// the input is not accepted, but we try to "box" it with some convenient modules
				
				// FLOAT => ANALOG 
				// boxing the input into a Smoother	
				if ((type == AudioDataType.FLOAT) && ((m_expectedType & AudioDataType.ANALOG) != 0))
				{
					auto smoother = new Smoother();
					smoother.input("IN").setData(audioData);
					setData(smoother.output("OUT").data);
				} else 
				
				// ANALOG => STEREO_ANALOG
				// boxing the input into a Doubler				
				if ((type == AudioDataType.ANALOG) && ((m_expectedType & AudioDataType.STEREO_ANALOG) != 0))
				{
					auto doubler = new Doubler();
					doubler.input("IN").setData(audioData);
					setData(doubler.output("OUT").data);
				} else 
				{
					throw new BadInputTypeError(owner.name, name, m_expectedType);
				}
			}			
		}		
	}
	
	protected
	{
		override bool canBeLinkedTo(Pin pin)
		{
			InputPin inputPin = cast(InputPin)pin;
			
			if (inputPin !is null)
			{
				// pin must not support data types not supported by this
				AudioDataTypeSet src = inputPin.m_expectedType;
				AudioDataTypeSet dest = m_expectedType;
				return (0 == (AudioDataType.ALL & ( src & (!dest))));
				
			} else return false; // linked pin should also be an InputPin
		}
	}
	
	public
	{
		this(AudioModule owner, char[] name, AudioDataTypeSet expectedType)
		{
			super(owner, name, null); // not connected yet
			m_expectedType = expectedType;
		}
		
		AudioDataTypeSet expectedType()
		{
			return m_expectedType;	
		}		
		
		void check()
		{
			auto d = data();
			if (d !is null)
			{
				if (!( d.type & m_expectedType ))
				{
					throw new BadInputTypeError(owner.name, name, m_expectedType);
				}
			} else
			{
				if (! m_expectedType & AudioDataType.UNASSIGNED)
				{
					throw new UnconnectedPinError(owner.name, name, m_expectedType);
				}
			}
		}
		
		void updateTo(AudioCycle targetCycle)
		{
			debug check();
			auto d = data();
			if (d !is null) 
			{
				d.updateTo(targetCycle);	
			}			
		}
		
		void connectTo(OutputPin outputPin)
		{
			setData(outputPin.data);			
		}
		
		void connectTo(AudioModule m)
		{
			connectTo(m.output());
		}
		
		void connectTo(float x)
		{
			connectTo((new Constant(x)).output());
		}
	}
}

final class OutputPin : Pin
{
	private
	{
		AudioDataType m_type;	
	}
		
	protected
	{
		override bool canBeLinkedTo(Pin pin)
		{
			OutputPin outputPin = cast(OutputPin)pin;
			
			if (outputPin !is null)
			{
				// pin data type must be the same as linked type				
				return (m_type == outputPin.m_type);
				
			} else return false; // linked pin should also be an Output pin
		}
	}
	
	public
	{
		this(AudioModule owner, char[] name, AudioDataType type)
		{
			super(owner, name, AudioDataFactory.create(type, owner));
			m_type = type;
		}	
		
		void setLastCycle(AudioCycle targetCycle)
		{
			data.lastCycle = targetCycle;
		}
	}
}